# -*- coding: utf-8 -*-
#
# Copyright 2011 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""A webclient backend that uses QtNetwork."""

import sys

# pylint: disable=E0611
from PyQt4.QtCore import (
    QBuffer,
    QCoreApplication,
    QUrl,
)
# pylint: enable=E0611
from PyQt4.QtNetwork import (
    QNetworkAccessManager,
    QNetworkProxy,
    QNetworkProxyFactory,
    QNetworkReply,
    QNetworkRequest,
)
from twisted.internet import defer

from ubuntu_sso.logger import setup_logging
from ubuntu_sso.utils.webclient.common import (
    BaseWebClient,
    HeaderDict,
    ProxyUnauthorizedError,
    Response,
    UnauthorizedError,
    WebClientError,
)
from ubuntu_sso.utils.webclient import gsettings

logger = setup_logging("ubuntu_sso.utils.webclient.qtnetwork")


class WebClientProxyFactory(QNetworkProxyFactory):
    """Proxy factory used by the web client.

    The proxy factory ensures that the correct proxy is used according to
    the request performed by the web client. For example, if the system has
    different proxies set for the http and the https the proxy factory will
    return the correct proxy according to the scheme of the request.
    """

    def __init__(self):
        """Create a new instance."""
        super(WebClientProxyFactory, self).__init__()
        self._proxies = dict(default=QNetworkProxy(QNetworkProxy.DefaultProxy))
        self._setup_proxies()

    def _get_proxy_for_settings(self, settings):
        """Return a proxy that uses the given settings."""
        if "username" in settings and "password" in settings:
            logger.debug('Setting auth %s:%s proxy',
                    settings.get("host", ""), settings.get("port", 0))
            proxy = QNetworkProxy(QNetworkProxy.HttpProxy,
                                  hostName=settings.get("host", ""),
                                  port=settings.get("port", 0),
                                  user=settings.get("username", ""),
                                  password=settings.get("password", ""))
        else:
            logger.debug('Using nonauth proxy %s:%s', settings.get("host", ""),
                    settings.get("port", 0))
            proxy = QNetworkProxy(QNetworkProxy.HttpProxy,
                                  hostName=settings.get("host", ""),
                                  port=settings.get("port", 0))
        return proxy

    def _setup_proxies(self):
        """Set the different proxies used in the system."""
        proxy_settings = gsettings.get_proxy_settings()
        for scheme, settings in proxy_settings.iteritems():
            self._proxies[scheme] = self._get_proxy_for_settings(settings)

    # ignore the complain since we have to use the Qt naming convention
    # pylint: disable=C0103
    def queryProxy(self, query):
        """Return the proxy servers to be used in order of preference."""
        if 'force' in self._proxies:
            return [self._proxies['force']]
        if self._proxies:
            scheme = str(query.protocolTag())
            if scheme in self._proxies:
                return [self._proxies[scheme]]
        # this function most return always at least one item, since we do not
        # have a proxy set for the query let's return the system default proxy
        return [self._proxies['default']]
    # pylint: enable=C0103

    def update_proxy(self, scheme, settings):
        """Allows to update the proxy to be used for a scheme."""
        self._proxies[scheme] = self._get_proxy_for_settings(settings)

    def force_use_proxy(self, settings):
        """Force to use the proxies in the given settings."""
        self._proxies['force'] = self._get_proxy_for_settings(settings)

    def find_proxy_schemes(self, proxy):
        """Return the scheme in wich the proxy is used."""
        schemes = []
        for scheme, scheme_proxy in self._proxies.iteritems():
            if scheme_proxy == proxy:
                schemes.append(scheme)
        return schemes


class WebClient(BaseWebClient):
    """A webclient with a qtnetwork backend."""

    def __init__(self, *args, **kwargs):
        """Initialize this instance."""
        super(WebClient, self).__init__(*args, **kwargs)
        self.nam = QNetworkAccessManager(QCoreApplication.instance())
        self.nam.finished.connect(self._handle_finished)
        self.nam.authenticationRequired.connect(self._handle_authentication)
        self.nam.proxyAuthenticationRequired.connect(
                                      self._handle_bad_proxy_authentication)
        self.nam.sslErrors.connect(self._handle_ssl_errors)
        self.replies = {}
        self.proxy_factory = None
        self.setup_proxy()

    def setup_proxy(self):
        """Setup the proxy settings if needed."""
        # QtNetwork knows how to use the system settings on both Win and Mac
        if sys.platform.startswith("linux"):
            self.proxy_factory = WebClientProxyFactory()
            self.nam.setProxyFactory(self.proxy_factory)
        else:
            QNetworkProxyFactory.setUseSystemConfiguration(True)

    def _perform_request(self, request, method, post_buffer):
        """Return a deferred that will be fired with a Response object."""
        d = defer.Deferred()
        if method == "GET":
            reply = self.nam.get(request)
        elif method == "HEAD":
            reply = self.nam.head(request)
        else:
            reply = self.nam.sendCustomRequest(request, method, post_buffer)
        # deal with auth errors in case we get
        # ProxyAuthenticationRequiredError
        d.addErrback(self._handle_proxy_authentication, method, reply,
                     post_buffer)
        self.replies[reply] = d
        return d

    @defer.inlineCallbacks
    def request(self, iri, method="GET", extra_headers=None,
                oauth_credentials=None, post_content=None):
        """Return a deferred that will be fired with a Response object."""
        uri = self.iri_to_uri(iri)
        request = QNetworkRequest(QUrl(uri))

        if extra_headers:
            headers = dict(extra_headers)
        else:
            headers = {}

        if oauth_credentials:
            timestamp = yield self.get_timestamp()
            oauth_headers = self.build_oauth_headers(method, uri,
                                                oauth_credentials, timestamp)
            headers.update(oauth_headers)

        for key, value in headers.iteritems():
            request.setRawHeader(key, value)

        post_buffer = QBuffer()
        post_buffer.setData(post_content)
        result = yield self._perform_request(request, method, post_buffer)
        defer.returnValue(result)

    def _handle_authentication(self, reply, authenticator):
        """The reply needs authentication."""
        authenticator.setUser(self.username)
        authenticator.setPassword(self.password)

    @defer.inlineCallbacks
    def _handle_bad_proxy_authentication(self, proxy, authenticator):
        """Handle when we have a bad proxy."""
        host = unicode(proxy.hostName()).encode('utf8')
        port = proxy.port()
        settings = dict(host=host, port=port)
        # HACK: work around for reported bugs: QTBUG-16728 and #QTBUG-14850
        got_creds = yield self.request_proxy_auth_credentials(host, True)
        if got_creds:
            logger.debug('Got proxy credentials from user.')
            settings.update({'username': self.proxy_username,
                             'password': self.proxy_password})
        if (settings and 'username' in settings
               and 'password' in settings):
            authenticator.setUser(self.proxy_username)
            authenticator.setPassword(self.proxy_password)
            schemes = self.proxy_factory.find_proxy_schemes(proxy)
            for scheme in schemes:
                self.proxy_factory.update_proxy(scheme, settings)

    @defer.inlineCallbacks
    def _handle_proxy_authentication(self, failure, method, reply,
            post_buffer=None, retry=False):
        """The reply needs authentication."""
        failure.trap(ProxyUnauthorizedError)
        current_proxy = self.nam.proxy()
        host = unicode(current_proxy.hostName()).encode('utf8')
        port = current_proxy.port()
        settings = dict(host=host, port=port)
        # we have to auth the request and retry it
        if not retry and self.proxy_username and self.proxy_password:
            settings.update({'username': self.proxy_username,
                             'password': self.proxy_password})
        if retry or self.proxy_username is None or self.proxy_password is None:
            got_creds = yield self.request_proxy_auth_credentials(host, retry)
            if got_creds:
                logger.debug('Got proxy credentials from user.')
                settings.update({'username': self.proxy_username,
                                 'password': self.proxy_password})
        if (settings and 'username' in settings
                and 'password' in settings):
            request = reply.request()
            scheme = str(request.url().scheme())
            self.proxy_factory.update_proxy(scheme, settings)
            # let's retry using the request
            result = yield self._perform_request(request, method, post_buffer)
            defer.returnValue(result)
        else:
            defer.returnValue(ProxyUnauthorizedError())

    def _handle_finished(self, reply):
        """The reply has finished processing."""
        assert reply in self.replies
        d = self.replies.pop(reply)
        error = reply.error()
        content = reply.readAll()
        if not error:
            headers = HeaderDict()
            for key, value in reply.rawHeaderPairs():
                headers[str(key)].append(str(value))
            response = Response(bytes(content), headers)
            d.callback(response)
        else:
            error_string = reply.errorString()
            logger.debug('_handle_finished error (%s,%s).', error,
                         error_string)
            if error == QNetworkReply.AuthenticationRequiredError:
                exception = UnauthorizedError(error_string, content)
            elif error == QNetworkReply.ProxyAuthenticationRequiredError:
                # we are going thru a proxy and we did not auth
                exception = ProxyUnauthorizedError(error_string, content)
            else:
                exception = WebClientError(error_string, content)
            d.errback(exception)

    @defer.inlineCallbacks
    def _handle_ssl_errors(self, reply, errors):
        """Handle the case in which we got an ssl error."""
        # ask the user if the cer should be trusted
        cert_pem = errors[0].certificate().toPem()
        trust_cert = yield self.request_ssl_cert_approval(cert_pem)
        if trust_cert:
            reply.ignoreSslErrors()

    def force_use_proxy(self, settings):
        """Setup this webclient to use the given proxy settings."""
        self.proxy_factory.force_use_proxy(settings)

    def shutdown(self):
        """Shut down all pending requests (if possible)."""
        self.nam.deleteLater()
