# -*- coding: utf-8 -*-
#
# Copyright 2011 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Store and retrieve certificates that have been pinned."""

import base64
import os
import tempfile

from ubuntu_sso import xdg_base_directory
from ubuntu_sso.logger import setup_logging

PINNED_CERT_LINE_FORMAT = ('%(domain)s:%(port)s\t%(hash_algorithm)s\t'
                           '%(fingerprint)s\t%(override_type)s\t'
                           '%(serial_number)s\n')
HTTPS_PORT = 443
HASH_ALGORITHM = ('sha1', 'OID.2.16.840.1.101.3.4.2.1')
DEFAULT_OVERRIDE_TYPE = 'MUT'
PINNED_CERT_DIR = os.path.join(xdg_base_directory.xdg_cache_home, 'sso',
                                                                  'ssl-certs')

logger = setup_logging("ubuntu_sso.utils.webclient.ssl_certs")


class PinnedSSLCertsVault(object):
    """Represents a vault that will store the pinned certs."""

    def __init__(self, certs_path=PINNED_CERT_DIR):
        """Create a new instance."""
        self.certs_path = certs_path
        if not os.path.exists(self.certs_path):
            os.makedirs(self.certs_path)

    def _get_pinned_certificate_line(self, cert):
        """Return the line of a pinned cert.

        The following format is used:
        Fields are separated by a tab character. Each line is terminated
        by a line feed character (UNIX format).

        1. domainname:port : port 443 for HTTPS (SSL)
        2. hash algorithm OID:
            SHA1-256: OID.2.16.840.1.101.3.4.2.1 (most used)
            SHA-384: OID.2.16.840.1.101.3.4.2.2
            SHA-512: OID.2.16.840.1.101.3.4.2.3
        3. Certificate fingerprint using previous hash algorithm
        4. One or more characters for override type:
            M : allow mismatches in the hostname
            U : allow untrusted certs (whether it's self signed cert or
                a missing or invalid issuer cert)
            T : allow errors in the validity time, for example, for expired or
                not yet valid certs
        5. Certificate's serial number and the issuer name as a base64
            encoded string

        more details can be found at:
        https://developer.mozilla.org/En/Cert_override.txt
        """
        cert_subject = cert.get_subject()
        serial_number = '%s@%s' % (cert.get_serial_number(), cert.get_issuer())
        details = dict(domain=cert_subject.CN,
                       port=HTTPS_PORT,
                       hash_algorithm=HASH_ALGORITHM[1],
                       fingerprint=cert.digest(HASH_ALGORITHM[0]),
                       override_type=DEFAULT_OVERRIDE_TYPE,
                       serial_number=base64.b64encode(serial_number))
        return PINNED_CERT_LINE_FORMAT % details

    def _get_cert_file_name(self, cert):
        """Return the file name to use with the cert."""
        fingerprint = cert.digest(HASH_ALGORITHM[0])
        return fingerprint.replace(':', '-')

    def store_certificate(self, cert):
        """Store a new certificate."""
        _, temp_path = tempfile.mkstemp(prefix='tmp-', dir=self.certs_path)
        with open(temp_path, 'wb') as fd:
            fd.write(self._get_pinned_certificate_line(cert))
        logger.debug('Temp file %s written', temp_path)
        fingerprint = self._get_cert_file_name(cert)
        os.rename(temp_path, os.path.join(self.certs_path, fingerprint))
        logger.debug('Cert with fingerprint %s added.', fingerprint)

    def is_certificate_pinned(self, cert):
        """Return if a certificate was already pinned."""
        fingerprint = self._get_cert_file_name(cert)
        return fingerprint in self.pinned_certificates

    @property
    def pinned_certificates(self):
        """Return all the fingerprints of the stored certs."""
        return [name for name in os.listdir(self.certs_path) if not
                name.startswith('tmp-')]
