# -*- coding: utf-8 -*-
#
# Copyright 2011 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the ssl certs operations."""

import os
import tempfile

from OpenSSL import crypto
from socket import gethostname
from twisted.internet import defer, reactor
from twisted.trial.unittest import TestCase

from ubuntu_sso.utils.webclient import ssl_certs


class PinnedSSLCertsVaultTestCase(TestCase):
    """Test the addition of a new pinned cert."""

    @defer.inlineCallbacks
    def setUp(self):
        """Set the diff tests."""
        yield super(PinnedSSLCertsVaultTestCase, self).setUp()
        self.cert_details = dict(organization='Canonical',
                                 common_name=gethostname(),
                                 locality_name='London',
                                 unit='Ubuntu One',
                                 country_name='UK',
                                 state_name='London',)
        self.cert = self._generate_cert(self.cert_details)

        self.certs_path = self.mktemp()
        os.mkdir(self.certs_path)
        self.vault = ssl_certs.PinnedSSLCertsVault(self.certs_path)

    def _generate_cert(self, cert_details):
        """Return a cert using the given details."""
        # create a key pair
        key = crypto.PKey()
        key.generate_key(crypto.TYPE_RSA, 1024)

        # create a self-signed cert
        cert = crypto.X509()
        cert.get_subject().C = cert_details['country_name']
        cert.get_subject().ST = cert_details['state_name']
        cert.get_subject().L = cert_details['locality_name']
        cert.get_subject().O = cert_details['organization']
        cert.get_subject().OU = cert_details['unit']
        cert.get_subject().CN = cert_details['common_name']
        cert.set_serial_number(1000)
        cert.gmtime_adj_notBefore(0)
        cert.gmtime_adj_notAfter(10 * 365 * 24 * 60 * 60)
        cert.set_issuer(cert.get_subject())
        cert.set_pubkey(key)
        cert.sign(key, 'sha1')

        return cert

    def test_missing_dir(self):
        """Test addind a pinned cert when the dir is missing."""
        os.rmdir(self.certs_path)
        self.vault = ssl_certs.PinnedSSLCertsVault(self.certs_path)
        self.assertTrue(os.path.exists(self.certs_path))

    def test_pinned_certificates(self):
        """Test getting the list of pinned certificates."""
        # create number of certs and ensure that the list is there.
        added = []

        for index in (1, 2, 3):
            cert_details = self.cert_details.copy()
            cert_details['common_name'] = \
                    cert_details['common_name'] + str(index)
            cert = self._generate_cert(cert_details)
            self.vault.store_certificate(cert)
            added.append(
                    cert.digest(ssl_certs.HASH_ALGORITHM[0]).replace(':', '-'))

        for fingerprint in added:
            self.assertIn(fingerprint, self.vault.pinned_certificates)

    def test_pinned_certificates_temp_files(self):
        """Test getting the list of pinned certificates."""
        # create a number of tmp files and assert that they are not returned.
        tempfiles = []
        added = []

        for index in (1, 2, 3):
            _, temp_path = tempfile.mkstemp(prefix='tmp-', dir=self.certs_path)
            with open(temp_path, 'w') as fd:
                fd.write("I'm a test!")
            tempfiles.append(temp_path)

        for index in (1, 2, 3):
            cert_details = self.cert_details.copy()
            cert_details['common_name'] = \
                    cert_details['common_name'] + str(index)
            cert = self._generate_cert(cert_details)
            self.vault.store_certificate(cert)
            added.append(
                    cert.digest(ssl_certs.HASH_ALGORITHM[0]).replace(':', '-'))

        for fingerprint in added:
            self.assertIn(fingerprint, self.vault.pinned_certificates)

        for temp in tempfiles:
            self.assertNotIn(temp, self.vault.pinned_certificates)

    def test_adding_certificated(self):
        """Test adding a new certificate."""
        self.vault.store_certificate(self.cert)
        fingerprint = self.cert.digest(
                ssl_certs.HASH_ALGORITHM[0]).replace(':', '-')
        self.assertIn(fingerprint, os.listdir(self.certs_path))

    def _assert_added(self):
        """Assert the addition with multiple threads."""
        vault = ssl_certs.PinnedSSLCertsVault(self.certs_path)
        vault.store_certificate(self.cert)
        fingerprint = self.cert.digest(
                ssl_certs.HASH_ALGORITHM[0]).replace(':', '-')
        self.assertIn(fingerprint, vault.pinned_certificates)
        self.assertIn(fingerprint, os.listdir(self.certs_path))

    def test_stepping_on_each_other(self):
        """Test using several vaults in threads."""
        # pylint: disable=E1101
        for _ in (1, 2, 3, 4, 5):
            reactor.callInThread(self._assert_added)
        # pylint: enable=E1101

    def test_get_cert_filename(self):
        """Assert that the filename is correct."""
        expected = self.cert.digest(
                ssl_certs.HASH_ALGORITHM[0]).replace(':', '-')
        # pylint: disable=W0212
        filename = self.vault._get_cert_file_name(self.cert)
        # pylint: enable=W0212
        self.assertEqual(expected, filename)
